package IO::Pty::Script;
use 5.008008;
use strict;
use warnings;
use IO::Handle;
use IO::Pty;
use IO::Select;
use Carp;
use POSIX;
use UNIVERSAL qw( isa );

require Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = ( qw{TIMEOUT DEFAULT_DEADLINE chats} );
our @EXPORT = qw( );
our $VERSION = '0.02';

use constant BUFFER_SIZE => 4096;
use constant DEBUG => 1;
use constant DEFAULT_DEADLINE => 4;
use constant TIMEOUT => 0.2;

sub new {
  my $class = shift;
  my $self = { @_ };
  bless $self, $class;
}

sub do_command {
  my ($pty, $STAT_RDR, $STAT_WTR, $command, @args) = @_;

  $pty->set_raw();
  defined(my $child = fork) or carp "Can't fork: $!";
  return if $child;
  close $STAT_RDR;
  $pty->make_slave_controlling_terminal();
  my $tty = $pty->slave();
  $tty->set_raw();
  close($pty);
  STDIN->fdopen($tty, "<");
  STDOUT->fdopen($tty, ">");
  STDERR->fdopen($tty, ">");
  close($tty);
  $| = 1;
  { exec $command, @args; }
  print $STAT_WTR $!+0; # Fuerza contexto numrico
}

{
  my $line = '';

  sub waitfor {
    my $pty = shift;
    my $sel = shift;
    my $regexp = shift;
    my $seconds = shift;
    my ($localline, $delayed, $nr, $parenthesis);

    alarm($seconds);
    eval {
      while ($nr = sysread($pty, $line, BUFFER_SIZE, length($line))) {
        last if $line =~ m{$regexp};
      }
    };
    alarm(0);
    # Leer si hay algo en la seudoterminal despues de TIMEOUT
    while ($sel->can_read(TIMEOUT)) {
      $nr = sysread($pty, $line, BUFFER_SIZE, length($line));
    }
    $line =~ m{($regexp)};
    if (defined($1)) {
      $localline = substr($line, 0, $+[0]);
      $line = substr($line, length($localline));
      $delayed = 0;
      no strict 'refs';
      $parenthesis = [ map { ${$_} } 2..$#+ ];
    }
    else {
      $localline = $line;
      $line = '';
      $delayed = 1;
    }
    my $is_eof = defined($nr) && ($nr == 0);
    return wantarray? ($localline, $line, $delayed, $parenthesis, $is_eof) : $localline;
  }
}

sub printpty {
  my $pty = shift;
  my $sel = shift;
  my $message = shift;
  do {} until $sel->can_write(TIMEOUT);
  syswrite($pty, $message);
}

# TODO
sub whilematch {
  my $self = shift;
  my @script = @{shift()};
  # No vale usar un hash porque el orden influye en las regexp perl
  my @questions = map { ($_%2)?():($script[$_]) } 0..$#script;
  my $questions;
  {
    local $"=')|(';
    $questions = "(@questions)";
  }
  {
    redo;
  }
}

sub chat {
  my $self = shift;
  my $command = $self->command() || 
                  carp "Error in chat: command argument missed";;
  my $deadline = $self->deadline() ||  DEFAULT_DEADLINE;

  carp "Error in chat: script argument missed" unless defined($self->script());
  my @script = @{$self->script()};
  my $action = $self->defaultaction() || sub {};
  my ($r, $s, $d, $p, $eof);

  my $pty = IO::Pty->new or carp "Can't make Pty: $!";
  my $sel = IO::Select->new();
  $sel->add($pty);
  my ($STAT_RDR, $STAT_WTR);
  pipe($STAT_RDR, $STAT_WTR) or carp "Cannot open pipe: $!";
  do_command($pty, $STAT_RDR, $STAT_WTR, $command);
  close $STAT_WTR;
  my $errno;
  my $errstatus = sysread($STAT_RDR, $errno, 256);
  carp "Cannot sync with child: $!" if not defined $errstatus;
  close $STAT_RDR;
  if ($errstatus) {
    $! = $errno+0;
    carp "Cannot exec(@ARGV): $!";
  }
  my $alarmhandler = sub { die; };
  local $SIG{ALRM} = $alarmhandler;
  while (@script) {
    my $question = shift @script;
    if (isa $question, "ARRAY") {
      $self->whilematch($question);
      next; # label?
    }
    my $answer   = shift @script;
    ($r, $s, $d, $p, $eof) = waitfor($pty, $sel, $question, $deadline);
    my $qaa = IO::Pty::Script::Answer->new(
                    pty => $pty, 
                    sel => $sel, 
                    script => \@script,
                    answer => $r, 
                    overpassed => $s, 
                    delayed => $d, 
                    parenthesis => $p,
                    eof => $eof, 
                    deadline => \$deadline
                 );
    if (isa $answer, "CODE") {
      $answer = $answer->($qaa);
    }
    elsif (isa $answer, "ARRAY") {
      my $func = shift @$answer || carp "Empty array of parameters";
      if (isa $func, "CODE") {
        my @params =  @$answer;
        $answer = $func->($qaa, @params);
      }
      else {
        $action->($qaa, @$answer);
      }
    }
    else { # $answer is a string
      $action->($qaa);
    }
    printpty($pty, $sel, $answer);
  }
  close($pty);
}

sub chats {
  while (@_) {
    my $self = shift;
    $self->chat();
  }
}

our $AUTOLOAD;
sub AUTOLOAD {
  my $self = shift;

  $AUTOLOAD =~ /.*::(\w+)/;
  my $subname = $1;
  carp "No such method $AUTOLOAD " unless (defined($subname));
  no strict 'refs';
  if (exists($self->{$subname})) {
    *{$AUTOLOAD} = sub { 
                     $_[0]->{$subname} = $_[1] if @_ >1; 
                     $_[0]->{$subname} 
                   };
    $self->{$subname} = $_[0] if @_;
    return $self->{$subname};
  }
  carp "No such method $AUTOLOAD";
}

sub DESTROY {
}

1;

#######################################################################
package IO::Pty::Script::Answer;
use Carp;
use strict;
our $AUTOLOAD;

sub new {
  my $class = shift;
  my $self = { @_ };
  bless $self, $class;
}

use overload q("") => \&strqa;
sub strqa {
  my $self = shift;

  my $r = $self->answer();
  my $s = $self->overpassed();
  my $d = $self->delayed();
  return <<"EOL";
<<r = '$r'
s = '$s'
d = '$d'>>
EOL
}

sub redirect {
  my ($src,$dst) = @_;
  my $buf = '';
  my $read = sysread($src, $buf, 1);
  if (defined $read && $read) {
    syswrite($dst,$buf,$read);
  }
  else { # EOF
    print STDERR "Nothing from $src";
    print "$read\n" if defined($read);
  }
  return $buf;
}

sub keyboard {
  my $self = shift;
  my $escape = shift;
  my $return_value = shift;

  my $char;
  my $pty = $self->pty();
  my $ws = $self->sel();
  my $rs = IO::Select->new();
  $rs->add(\*STDIN, $pty);
WHILE_NOT_ESCAPE_OR_EOF:  
  { # infinite loop
    my @ready = $rs->can_read(IO::Pty::Script::TIMEOUT);
    if (@ready) {
      @ready = reverse @ready if (@ready >1) and ($ready[0] != $pty);
      if ($ready[0] == $pty) {
        my $read = sysread($pty, $char, 1);
        if (defined $read && $read) {
          syswrite(STDOUT,$char,$read);
        }
        else { # EOF
          last WHILE_NOT_ESCAPE_OR_EOF 
        }
      }
      elsif ($ws->can_write(IO::Pty::Script::TIMEOUT)) { # Algo en STDIN
        my $read = sysread(STDIN, $char, 1);
        last WHILE_NOT_ESCAPE_OR_EOF if $char eq $escape;
        if (defined $read && $read) {
          syswrite($pty,$char,$read);
        }
        else {
          last WHILE_NOT_ESCAPE_OR_EOF 
        }
      }
    }
    redo;
  }
  return $return_value;
}

sub deadline {
  ${$_[0]->{deadline}} = $_[1] if @_ >1;
  ${$_[0]->{deadline}};
}

sub parenthesis {
  return @{$_[0]->{parenthesis}};
}

sub can_read {
  my $self = shift;
  my $deadline = shift;
  my $sel = $self->{sel};

  return $sel->can_read($deadline);
}

sub can_write {
  my $self = shift;
  my $deadline = shift;
  my $sel = $self->{sel};

  return $sel->can_write($deadline);
}

sub AUTOLOAD {
  my $self = shift;

  $AUTOLOAD =~ /.*::(\w+)/;
  my $subname = $1;
  carp "No such method $AUTOLOAD " unless (defined($subname));
  no strict 'refs';
  if (exists($self->{$subname})) {
    *{$AUTOLOAD} = sub { 
                     $_[0]->{$subname} = $_[1] if @_ >1; 
                     $_[0]->{$subname} 
                   };
    $self->{$subname} = $_[0] if @_;
    return $self->{$subname};
  }
  carp "No such method $AUTOLOAD";
}

sub DESTROY {
}

1;
__END__

=head1 NAME

IO::Pty::Script - Perl extension for 

=head1 SYNOPSIS

  use IO::Pty::Script;

=head1 DESCRIPTION


=head2 EXPORT



=head1 SEE ALSO

=head1 AUTHOR

Casiano Rodriguez-Leon, E<lt>casiano@ull.esE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2006 by Lenguajes y Herramientas de Programacion

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
